@@ -11,19 +11,89 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <stdio.h>
-#include <unistd.h>
+#ifdef HAVE_UNISTD_H
+#  include <unistd.h>
+#endif
 #include <errno.h>
 #include <string.h>
 #include <fcntl.h>
 
 #include <glib.h>
-#include "linc-debug.h"
+
+#include "linc-compat.h"
 #include "linc-private.h"
 
+#if defined (G_OS_WIN32) && defined (CONNECTION_DEBUG)
+
+static char *
+fd_mask (int mask)
+{
+  static char buf[100];
+  char *bufp = buf;
+
+  if (mask == 0)
+    return "";
+
+#define BIT(n) if (mask & FD_##n) bufp += sprintf (bufp, "%s" #n, (bufp>buf ? "|" : ""))
+
+  BIT (READ);
+  BIT (WRITE);
+  BIT (OOB);
+  BIT (ACCEPT);
+  BIT (CONNECT);
+  BIT (CLOSE);
+  BIT (QOS);
+  BIT (GROUP_QOS);
+  BIT (ROUTING_INTERFACE_CHANGE);
+  BIT (ADDRESS_LIST_CHANGE);
+  
+#undef BIT
+
+  return buf;
+}
+
+#endif					   
+
 static gboolean 
 link_source_prepare (GSource *source,
 		     gint    *timeout)
 {
+#ifdef G_OS_WIN32
+	LinkUnixWatch *watch = (LinkUnixWatch *) source;
+	int event_mask = 0;
+
+#if 0
+	if (watch->condition & G_IO_IN) {
+		u_long nread;
+		if (ioctlsocket (watch->socket, FIONREAD, &nread) == 0) {
+			d_printf ("FIONREAD %d: %d\n", watch->socket, nread);
+			if (nread > 0) {
+				watch->pollfd.revents = G_IO_IN;
+				return TRUE;
+			}
+		}
+	}
+#endif
+
+	if (watch->condition & G_IO_IN)
+		event_mask |= (FD_READ | FD_ACCEPT | FD_CLOSE);
+	if (watch->condition & G_IO_OUT)
+		event_mask |= (FD_WRITE /*| FD_CONNECT*/);
+	if (watch->condition & G_IO_HUP)
+		event_mask |= FD_CLOSE;
+
+	if (watch->event_mask != event_mask) {
+		d_printf ("WSAEventSelect(%d, %#x, {%s})\n",
+			  watch->socket, watch->pollfd.fd, fd_mask (event_mask));
+		if (WSAEventSelect (watch->socket, (HANDLE) watch->pollfd.fd,
+				    event_mask) == SOCKET_ERROR)
+			d_printf ("WSAEventSelect() failed: %s\n",
+				  link_strerror (WSAGetLastError ()));
+		else
+			watch->event_mask = event_mask;
+	}
+#endif
+
 	*timeout = -1;
 
 	return FALSE;
@@ -49,6 +119,31 @@ link_source_dispatch (GSource    *source,
 	if (!callback)
 		g_error ("No callback");
   
+#ifdef G_OS_WIN32
+	if (watch->pollfd.revents & watch->condition) {
+		WSANETWORKEVENTS events;
+
+		d_printf ("WSAEnumNetworkEvents (%d, %#x)\n",
+			  watch->socket, watch->pollfd.fd);
+		if (WSAEnumNetworkEvents (watch->socket,
+					  (HANDLE) watch->pollfd.fd,
+					  &events) == SOCKET_ERROR)
+			d_printf ("WSAEnumNetworkEvents failed: %s\n",
+				  link_strerror (WSAGetLastError ()));
+		else {
+			d_printf ("WSAEnumNetworkEvents: lNetworkEvents={%s}\n",
+				  fd_mask (events.lNetworkEvents));
+			watch->pollfd.revents = 0;
+			if (events.lNetworkEvents & (FD_READ | FD_ACCEPT | FD_CLOSE))
+				watch->pollfd.revents |= G_IO_IN;
+			if (events.lNetworkEvents & (FD_WRITE /*| FD_CONNECT*/))
+				watch->pollfd.revents |= G_IO_OUT;
+			if (events.lNetworkEvents & (FD_CLOSE))
+				watch->pollfd.revents |= G_IO_HUP;
+		}
+	}
+#endif
+
 	func = (GIOFunc) callback;
 
 	return (*func) (watch->channel,
@@ -62,7 +157,7 @@ link_source_finalize (GSource *source)
 	d_printf ("Finalize source %p\n", source);
 }
 
-GSourceFuncs link_source_watch_funcs = {
+static GSourceFuncs link_source_watch_funcs = {
 	link_source_prepare,
 	link_source_check,
 	link_source_dispatch,
@@ -117,7 +212,14 @@ link_source_create_watch (GMainContext *context,
 			       sizeof (LinkUnixWatch));
 	watch = (LinkUnixWatch *) source;
 
+#ifdef G_OS_WIN32
+	watch->pollfd.fd = (int) WSACreateEvent ();
+	d_printf ("WSACreateEvent(): %#x\n", watch->pollfd.fd);
+	watch->socket = fd;
+	watch->event_mask = 0;
+#else
 	watch->pollfd.fd = fd;
+#endif
 	watch->channel   = opt_channel;
 	watch->condition = condition;
 	watch->callback  = func;
@@ -160,7 +262,8 @@ link_io_add_watch_fd (int          fd,
 			 condition, func, user_data);
 		
 		w->main_source = link_source_create_watch
-			(NULL, fd, NULL, condition, func, user_data);
+			(NULL, fd, NULL,
+			 condition, func, user_data);
 	}
 
 	return w;
@@ -171,6 +274,12 @@ link_watch_unlisten (LinkWatch *w)
 {
 	if (w->main_source) {
 		link_source_set_condition (w->main_source, 0);
+#ifdef G_OS_WIN32
+		d_printf ("CloseHandle(%#x)\n",
+			  ((LinkUnixWatch *) w->main_source)->pollfd.fd);
+		if (!CloseHandle ((HANDLE) ((LinkUnixWatch *) w->main_source)->pollfd.fd))
+			d_printf ("CloseHandle failed: %ld\n", GetLastError ());
+#endif
 		g_source_destroy (w->main_source);
 		g_source_unref   (w->main_source);
 		w->main_source = NULL;
@@ -178,6 +287,12 @@ link_watch_unlisten (LinkWatch *w)
 
 	if (w->link_source) {
 		link_source_set_condition (w->link_source, 0);
+#ifdef G_OS_WIN32
+		d_printf ("CloseHandle(%#x)\n",
+			  ((LinkUnixWatch *) w->link_source)->pollfd.fd);
+		if (!CloseHandle ((HANDLE) ((LinkUnixWatch *) w->link_source)->pollfd.fd))
+			d_printf ("CloseHandle failed: %ld\n", GetLastError ());
+#endif
 		g_source_destroy (w->link_source);
 		g_source_unref   (w->link_source);	
 		w->link_source = NULL;
